#ifndef PATCHER_H
#define PATCHER_H

#include <windows.h>

#define EXPECTED_OP2_ADDR 0x00400000
#define EXPECTED_OP2_NAME NULL // Calling GetModuleHandle(NULL) from a DLL returns the handle of the parent process


class _Patch;
class Patcher;


class Patcher
{
	// Static functions
public:

	// These functions rewrite arbitrary code or data
	static _Patch* Patch(void *address, void *newBytesBuffer, void *expectedBytesBuffer, unsigned int sizeOfBytesBuffer, bool onlyQueue);
	static _Patch* Patch(void *address, void *newBytesBuffer, unsigned int sizeOfBytesBuffer, bool onlyQueue);
	template <typename PatchType>
		static _Patch* Patch(void *address, PatchType newBytes, PatchType expectedBytes, bool onlyQueue);
	template <typename PatchType>
		static _Patch* Patch(void *address, PatchType newBytes, bool onlyQueue);


	// These functions rewrite/insert a JUMP or CALL instruction
	static _Patch* PatchFunction(void *functionAddress, void *newFunctionAddress, bool onlyQueue);
	static _Patch* PatchFunctionCall(void *functionCallAddress, void *newFunctionAddress, bool onlyQueue);


	// These functions patch a function pointer in a virtual function table
	static _Patch* Patcher::PatchFunctionVirtual(void *vtblAddress, void *functionAddress, void *newFunctionAddress, bool onlyQueue);
	template <class Class>
		static _Patch* PatchFunctionVirtual(void *functionAddress, void *newFunctionAddress, bool onlyQueue);
	template <class Class> // Use this if the class has no default constructor
		static _Patch* PatchFunctionVirtual(Class &classObj, void *functionAddress, void *newFunctionAddress, bool onlyQueue);

	static _Patch* Patcher::PatchFunctionVirtual(void *vtblAddress, int vtblEntryIndex, void *newFunctionAddress, bool onlyQueue);
	template <class Class>
		static _Patch* PatchFunctionVirtual(int vtblEntryIndex, void *newFunctionAddress, bool onlyQueue);
	template <class Class> // Use this if the class has no default constructor
		static _Patch* PatchFunctionVirtual(Class &classObj, int vtblEntryIndex, void *newFunctionAddress, bool onlyQueue);


	// This function patches all references to a global variable/object (provided they are in the module's base relocation table)
	static _Patch* Patcher::PatchGlobalReferences(void *oldGlobalAddress, void *newGlobalAddress, bool onlyQueue);


	// These functions mass patch/unpatch; mass patch is only useful for queued patches
	static int DoPatchAll();
	static int DoUnpatchAll(bool doDelete, bool force = false);


	// Pre-allocates memory for the next patch object; only used with the #define functions
	static _Patch* GetNextPatchAddress();

private:
	static void AppendPatchToList(_Patch *patch);
	static void RemovePatchFromList(_Patch *patch, bool forceUnpatch = false);

	static void AllocateNextPatch();

	static DWORD InitModuleBase(char module[] = EXPECTED_OP2_NAME);

	// Static variables
private:
	static _Patch *patchListHead;					// Latest applied patch. List is sorted newest -> oldest
	static int numPatches;

	static _Patch *nextPatchAddress;				// Next allocated block of memory to store a new patch in

	static DWORD imageBase;

public:
	// These are for the #define functions
	static void *_mbrFunc;
	static void *_newMbrFunc;

public:
	// Friend classes
	friend class _Patch;

	// Friend functions
	friend inline DWORD OP2Addr(DWORD address);		// Fixes provided address against OP2's expected address and its load address (VA to VA)
	friend inline void* OP2Addr(void *address);
	friend inline DWORD _OP2Addr(DWORD address);	// Simplified version of OP2Addr() for use in code hooks or loops
	friend inline void* _OP2Addr(void *address);
	friend DWORD OP2RVAToVA(DWORD relAddress);		// Same as OP2Addr, but provided address is assumed to be a relative offset (RVA to VA)
	friend inline void* OP2RVAToVA(void *relAddress);
};


// Templates and inlined functions must be declared and defined in the same file

template <typename PatchType>
_Patch* Patcher::Patch(void *address, PatchType newBytes, bool onlyQueue)
{
	return Patcher::Patch(address, &newBytes, NULL, sizeof(PatchType), onlyQueue); 
}


template <typename PatchType>
_Patch* Patcher::Patch(void *address, PatchType newBytes, PatchType expectedBytes, bool onlyQueue)
{
	return Patcher::Patch(address, &newBytes, &expectedBytes, sizeof(PatchType), onlyQueue); 
}


template <class Class>
_Patch* Patcher::PatchFunctionVirtual(void *functionAddress, void *newFunctionAddress, bool onlyQueue)
{
	Class classObj; // Dummy for getting the vtbl pointer, doesn't work if the class has no default constructor
	return Patcher::PatchFunctionVirtual<Class>(classObj, functionAddress, newFunctionAddress, onlyQueue);
}


template <class Class>
_Patch* Patcher::PatchFunctionVirtual(Class &classObj, void *functionAddress, void *newFunctionAddress, bool onlyQueue)
{
	void **vtbl = (void**&)classObj;
	return Patcher::PatchFunctionVirtual((void*)vtbl, functionAddress, newFunctionAddress, onlyQueue);
}


template <class Class>
_Patch* Patcher::PatchFunctionVirtual(int vtblEntryIndex, void *newFunctionAddress, bool onlyQueue)
{
	Class classObj; // Dummy for getting the vtbl pointer, doesn't work if the class has no default constructor
	return Patcher::PatchFunctionVirtual<Class>(classObj, vtblEntryIndex, newFunctionAddress, onlyQueue);
}


template <class Class>
_Patch* Patcher::PatchFunctionVirtual(Class &classObj, int vtblEntryIndex, void *newFunctionAddress, bool onlyQueue)
{
	void **vtbl = (void**&)classObj;
	return Patcher::PatchFunctionVirtual(&vtbl[vtblEntryIndex], vtblEntryIndex, newFunctionAddress, onlyQueue);
}


// Fixes provided address against OP2's expected address and its load address (VA to VA)
inline DWORD OP2Addr(DWORD address) { return OP2RVAToVA(address - EXPECTED_OP2_ADDR); }
inline void* OP2Addr(void *address) { return (void*)(OP2Addr((DWORD)address)); }

// Simplified inline versions of OP2Addr() for use in loops or code hooks. These assume imageBase is already set!
inline DWORD _OP2Addr(DWORD address) { return (DWORD)(address - EXPECTED_OP2_ADDR + Patcher::imageBase); }
inline void* _OP2Addr(void *address) { return (void*)(_OP2Addr((DWORD)address)); }

// void* input/output for OP2RVAToVA(). Input address must be relative offset
inline void* OP2RVAToVA(void *relAddress) { return (void*)OP2RVAToVA((DWORD)relAddress); }


// Do not instantiate this class yourself. Use the Patcher class to do so.
class _Patch
{
	// Member functions
public:
	_Patch();
	_Patch(void *address, void *newBytes, void *expectedBytes, unsigned int sizeOfBytes, bool doApply);
	_Patch& operator= (const _Patch &source);
	~_Patch();

	int Patch(bool repatch = false);
	int Unpatch(bool doDelete, bool force = false);

	// Use these if you want a copy of the original unmodified bytes this patch overwrites
	void* CopyOriginalBytes(void *buffer, unsigned int sizeOfBuffer);
	template <typename ReturnType>
		ReturnType GetOriginalBytes();

	bool Applied();
	bool Valid();

	void SetPermanent(bool permanent);
	void SetLeaveMemoryUnprotected(bool leaveUnprotected);

	_Patch* GetNext();
	_Patch* GetPrevious();

	// Member variables
private:
	_Patch *next;					// Points towards older patches
	_Patch *previous;				// Points towards newer patches

	void *address;
	unsigned int patchSize;
	unsigned char *oldBytesBuffer;
	DWORD oldPageProtection;
	unsigned char *newBytesBuffer;

	bool isApplied;
	bool isPermanent;				// If the patch is set to permanent before being applied, the object will be deleted upon application;
									// if set to be permanent after application, the object will be preserved
	bool leaveMemoryUnprotected;
	bool invalid;

	// Friend classes
	friend class Patcher;
};


template <typename ReturnType>
ReturnType _Patch::GetOriginalBytes()
{
	ReturnType returnedValue;
	_Patch::CopyOriginalBytes(&returnedValue, sizeof(ReturnType));
	return returnedValue;
}



#define _GetFuncPtr(func, ptr) do { __asm mov eax, func __asm mov ptr, eax } while(0)

// These "functions" are used as ex. Patcher::PatchMemberFunction(Class::Function, ...);
//  They are used when you need to patch a class member function by directly referencing
//  Class::Function. Usually only necessary when it is a virtual function or you are not
//  using MSVC; otherwise you can get a plain void* pointer with &(void*&)Class::Function
// A caveat is that, if you use these inside conditional statements, enclosing the conditional body
//  within brackets is not optional, despite it looking like only one line of written code.

// Patcher::PatchVirtualFunction
#define PatchVirtualFunction(Class, classFunctionToPatch, newFunction, onlyQueue)					\
	GetNextPatchAddress();																			\
	_GetFuncPtr(classFunctionToPatch, Patcher::_mbrFunc);											\
	_GetFuncPtr(newFunction, Patcher::_newMbrFunc);													\
	Patcher::PatchFunctionVirtual<Class>(Patcher::_mbrFunc, Patcher::_newMbrFunc, onlyQueue)

// Patcher::PatchObjectVirtualFunction
#define PatchObjectVirtualFunction(obj, Class, classFunctionToPatch, newFunction, onlyQueue)		\
	GetNextPatchAddress();																			\
	_GetFuncPtr(classFunctionToPatch, Patcher::_mbrFunc);											\
	_GetFuncPtr(newFunction, Patcher::_newMbrFunc);													\
	Patcher::PatchFunctionVirtual<Class>(obj, Patcher::_mbrFunc, Patcher::_newMbrFunc, onlyQueue)

// Patcher::PatchMemberFunction
#define PatchMemberFunction(classFunctionToPatch, newFunction, onlyQueue)							\
	GetNextPatchAddress();																			\
	_GetFuncPtr(classFunctionToPatch, Patcher::_mbrFunc);											\
	_GetFuncPtr(newFunction, Patcher::_newMbrFunc);													\											\
	Patcher::PatchFunction(Patcher::_mbrFunc, Patcher::_newMbrFunc, onlyQueue)

// Patcher::PatchReferencedVirtualFunction
// To use this you need to make a struct filled with function pointers that represents the vftable you want to hook
#define PatchReferencedVirtualFunction(referencedFunctionToPatch, newFunction, onlyQueue)			\
	GetNextPatchAddress();																			\
	_GetFuncPtr(newFunction, Patcher::_newMbrFunc);													\
	Patcher::Patch(&referencedFunctionToPatch, &Patcher::_newMbrFunc, sizeof(void*), onlyQueue)

// Ultra-nasty workaround for the compiler not wanting to treat Class::Function as a pointer to the function.
//  In MSVC there is &(void*&)Class::Function to get a plain void* pointer to a class member function, but if
//  you try it on a virtual function, you may get a compiler-generated thunk that it makes specially for that
//  case. You also cannot pass Class::Function as a class argument without taking some kind of pointer to it.

#endif